
#include "dict.h"


DictNode::DictNode() {
        letter_flags = 0;
        no_hooks = 0;
        hook_alloc = 0;
        min_is_word = -1;
        parent = NULL;
}



// Prints the word represented by this node

void DictNode::print_word(ostream &where) {

        int e;

        if (NULL == parent)
                return;

        parent->print_word(where);

        for (e='a'; e<='z'; ++e) {

                if (parent->return_child(e) == this) {
                        where << char(e);
                        break;
                }

        }

}

// returns the DictNode corresponding to this letter, or NULL if none

DictNode *DictNode::return_child(int letter) {

        if (letter_flags & (1 << (letter - 'a'))) {

                int e, f, g;

                for (e=0, f=0, g=letter_flags; e<(letter-'a'); ++e) {

                        if (g & 1)
                                ++f;
                        g >>= 1;

                }

                return children[f];

        }

        return NULL;

}

// adds a child corresponding to the given letter, if none exists

DictNode *DictNode::add_child(int letter) {

        int e, f, g;

        if (letter < 'a' || letter > 'z')
                return NULL;

        if (letter_flags & (1 << (letter - 'a')))
                return return_child(letter);

        if (no_hooks == hook_alloc) {

                DictNode **b;

                hook_alloc += 4;

                if (hook_alloc > 30) {
                        cout << "Sanity check failure: hook_alloc exceeds 30 in node ";
                        print_word(cout);
                        cout << endl;
                        exit(-1);
                }

                b = new DictNode * [hook_alloc];

                if (NULL == b) {
                        cout << "Warning: error allocating memory loading dictionary!\n";
                        exit(-1);
                }

                for (e=0, f=0, g=letter_flags; e<(letter-'a'); ++e) {
                        if (g & 1)
                                f++;
                        g >>= 1;
                }
                if (f)
                        memcpy(b, children, sizeof(DictNode *) * f);

                b[f] = new DictNode();
                if (b[f] == NULL) {
                        cerr << "\nAbort! Error creating dictionary node!" << endl;
                        exit(-1);
                }
                b[f]->parent = this;
                b[f]->hook_alloc = 0;

                if (no_hooks - f)
                        memcpy(b + f + 1, children + f, sizeof(DictNode *) * (no_hooks - f));

                if (4 != hook_alloc)
                                delete [] children;

                children = b;

                no_hooks++;
                letter_flags |= (1 << (letter - 'a'));

                return children[f];

        }


        for (e=0, f=0, g=letter_flags; e<(letter-'a'); ++e) {
                if (g & 1)
                        f++;
                g >>= 1;
        }

        if (no_hooks - f)
                memmove(children + f, children + f + 1, sizeof(DictNode *) * (no_hooks - f));

        children[f] = new DictNode();
        if (children[f] == NULL) {
                cerr << "\nAbort! Error creating dictionary node!" << endl;
                exit(-1);
        }
        children[f]->parent = this;
        children[f]->hook_alloc = 0;

        no_hooks++;
        letter_flags |= (1 << (letter - 'a'));

        return children[f];

}

int DictNode::find_min_is_word() {

        int e, min;

        if (0 == min_is_word)
                return min_is_word;

        min = 99;
        for (e=0; e<no_hooks; ++e) {

                int f;

                f = children[e]->find_min_is_word();

                if (f < min) {
                        min = f;
                }

        }

        min_is_word = min+1;

        return min_is_word;

}

void DictNode::delete_node() {

        int e;

        for (e=0; e<no_hooks; ++e) {
                children[e]->delete_node();
        }
        for (e=0; e<no_hooks; ++e) {
                delete children[e];
        }
        delete [] children;

}

void Dictionary::load_dictionary(ifstream &in) {

        nwords = 0;

        while (!in.eof()) {

                int e;
                char word[64];
                DictNode *d;

                in >> word;

//                cout << word << endl;

                e=0;
                d=&top;
                while (word[e] && word[e] != '\n' && word[e] != '\r') {

                        int l;

                        l = tolower(word[e]);
                        if (l < 'a' || l > 'z') {
                                cout << "Warning: oddball dictionary entry: " << word << endl;
                                e = 0;
                                break;
                        }
                        if (NULL == d) {
                                e = 0;
                                break;
                        }

                        d = d->add_child(tolower(word[e]));    // won't create children that already exist

                        ++e;
                }

                if (e) {
                        d->min_is_word = 0;     // this is a word!
                        ++nwords;
                }

        }

        // now to navigate the tree, setting min_is_word for each node
        // not already set

        top.find_min_is_word();

}

void Dictionary::delete_dictionary() {

        top.delete_node();

        nwords = 0;

}

bool Dictionary::is_word(char *w) {

        DictNode *d;
        int e;

        d = &top;

        e = 0;
        while (w[e]) {
                d = d->return_child(w[e]);
                if (NULL == d)
                        return false;
                ++e;
        }
        if (!d->min_is_word)
                return true;

        return false;

}


// end dict.cxx
