
/*

        encodel.c

        Encodes the word list

        Chris Street, 2002

*/

#include <fstream.h>
#include <stdio.h>

#include "dict.h"

void write_bits(FILE *f, int value, int nbits) {

        static int outbits = 0;
        static int countbts = 0;
        static int bitcount = 0;

        /*

        	The heart and soul of Bit Output for bit-oriented I/O.
        	Writes out the data value to the specified file in nbits number
        	of bits (nbits <= 32.) Will encode the bits using the password
        	hash if we have one.
        	
                writes out nbits to file
                special options:
                nbits == -1 ---> reinitialize and return (call before writing to a new file)
                nbits == -2 ---> flush any pending bits and return

                this outputs multibyte values in little-endian format

        */

        if (nbits == -1) {
                outbits = 0;
                countbts = 0;
                bitcount = 0;
                return;
        }
        if (nbits == -2) {
                if (countbts) {
                        fputc(outbits, f);
                        countbts = 0;
                }
                return;
        }

        /* be sure only the low "nbits" of value are set */
        if (nbits != 32)
                value &= ((1 << nbits) - 1);

        outbits |= (value << countbts);

        countbts += nbits;
        bitcount += nbits;

        while (countbts >= 8) {
                fputc(outbits & 0xff, f);
                outbits >>= 8;
                countbts -= 8;
        }

}

void write_node(FILE *f, DictNode *n) {

        if (n->min_is_word)
                write_bits(f, 1, 1);
        else
                write_bits(f, 0, 1);

        int e, c;

        // determine number of children

        for (e='a', c = 0; e<='z'; ++e) {
                DictNode *ch;
                ch = n->return_child(e);
                if (ch != NULL)
                        c++;
        }

        write_bits(f, c, 5);

        for (e='a'; e<='z'; ++e) {
                DictNode *ch;
                ch = n->return_child(e);
                if (NULL != ch) {
                        write_bits(f, e - 'a', 5);
                        write_node(f, ch);
                }
        }

}

int main(int argc, char **argv) {

        FILE *o;
        ifstream i;
        Dictionary dict;

        if (argc < 3) {
                cerr << " Proper call: encodel [word list] [output]" << endl;
                return -1;
        }

        i.open(argv[1]);
        o = fopen(argv[2], "wb");

        dict.load_dictionary(i);

        // now, let's go!

        write_bits(o, 0, -1);   // initialize bit i/o stream

        DictNode node = dict.top;
        write_node(o, &node);

        fclose(o);
        i.close();

        return 0;

}


// end encodel.cxx

